;******************************************************************************
;   RX A7105 FlySky protocol PIC16F1824 6CH                                   *
;   PPM output on channel 6 option                                                                          *
;******************************************************************************
;                                                                             *
;    Filename:        xxx.asm                                                 *
;    Date:                                                                    *
;    File Version:                                                            *
;                                                                             *
;    Author:    Thierry Pbayle                                                              *
;    Company:                                                                 *
;                                                                             *
;                                                                             *
;******************************************************************************
;                                                                             *
;    Files Required: P16F1827.INC                                             *
;                                                                             *
;******************************************************************************
;                                                                             *
;    Notes:                                                                   *
;                                                                             *
;******************************************************************************
;                                                                             *
;    Revision History:                                                        *
;                                                                             *
;******************************************************************************

	#define RELEASE	;comment for debug (disable LED, MOTOR, ...)

	list		p=16f1824      ; list directive to define processor
	#include	<p16f1824.inc> ; processor specific variable definitions

;------------------------------------------------------------------------------
;
; CONFIGURATION WORD SETUP
;
; The 'CONFIG' directive is used to embed the configuration word within the 
; .asm file. The lables following the directive are located in the respective 
; .inc file.  See the data sheet for additional information on configuration 
; word settings.
;
;------------------------------------------------------------------------------    

#ifdef	RELEASE
    __CONFIG _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_OFF & _MCLRE_OFF & _CP_OFF & _CPD_OFF & _BOREN_OFF & _CLKOUTEN_OFF & _IESO_OFF & _FCMEN_OFF
    __CONFIG _CONFIG2, _WRT_OFF & _PLLEN_OFF & _STVREN_OFF & _BORV_19 & _LVP_OFF
#else
    __CONFIG _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_OFF & _MCLRE_ON & _CP_OFF & _CPD_OFF & _BOREN_OFF & _CLKOUTEN_OFF & _IESO_OFF & _FCMEN_OFF
    __CONFIG _CONFIG2, _WRT_OFF & _PLLEN_OFF & _STVREN_OFF & _BORV_19 & _LVP_OFF
#endif

;******************************************************************************
; CONSTANTES
;******************************************************************************

;LED
#define LED             PORTA, 1
#define LED_TRIS        TRISA, 1

;com spi A7105
#define CS_SPI          PORTA,4		;chip select SPI
#define SCK_SPI         PORTA,5		;serial clock SPI
#define	SDIO_SPI        PORTC,4     ;serial data input/output SPI
#define	GIO             PORTA,3		;GIO output A7105

#define CS_SPI_TRIS     TRISA,4		;chip select SPI
#define SCK_SPI_TRIS    TRISA,5		;serial clock SPI
#define	SDIO_SPI_TRIS   TRISC,4     ;serial data input/output SPI
#define	GIO_TRIS        TRISA,3		;GIO output A7105

;motor, aileron, elevator, rudder outputs
#define MOTOR   		PORTA, 0
#define AIL1      		PORTA, 2
#define AIL2        	PORTC, 0
#define ELE1        	PORTC, 1
#define ELE2        	PORTC, 2
#define RUD1        	PORTC, 3
#define RUD2        	PORTC, 5

#define MOTOR_TRIS   	TRISA, 0
#define AIL1_TRIS      	TRISA, 2
#define AIL2_TRIS       TRISC, 0
#define ELE1_TRIS       TRISC, 1
#define ELE2_TRIS       TRISC, 2
#define RUD1_TRIS       TRISC, 3
#define RUD2_TRIS		TRISC, 5

;flag aileron, elevator, rudder direction
#define ailDir 		direction, 0
#define eleDir 		direction, 1
#define rudDir 		direction, 2

;******************************************************************************
; MACRO
;******************************************************************************

;compare w and register, skip if equal
cpfseq MACRO reg
	subwf	reg, w
	btfss	STATUS, Z
	ENDM
	
;move register into register through W
;register1 -> W -> register2
movff MACRO reg1, reg2
    movf	reg1, w
    movwf   reg2
    ENDM

;move literal into register through W
;literal -> W -> register
movlwf MACRO valeur, reg
    movlw   valeur
    movwf   reg
    ENDM

initFSR0 macro adrReg
	banksel	FSR0H
    movlw   high adrReg
    movwf   FSR0H
    movlw   low adrReg
    movwf   FSR0L
    endm

initFSR1 macro adrReg
	banksel	FSR1H
    movlw   high adrReg
    movwf   FSR1H
    movlw   low adrReg
    movwf   FSR1L
    endm

;enable/disable SPI (CS)
SPI_ON MACRO
	banksel	PORTA
	bcf		CS_SPI					;enable SPI A7105
	ENDM

SPI_OFF	MACRO
	banksel	PORTA
	bsf 	CS_SPI					;disable SPI A7105
	ENDM

;one clock pulse for spi com
SCK_SPI_PULSE MACRO		;one clock (rising edge then falling edge)
	banksel	PORTA
    bsf		SCK_SPI
    bcf		SCK_SPI
    ENDM

;SPI write MACRO (adr = register, val = register)
spi_write_fadr_fval MACRO fadr, fval
    SPI_ON
	movff	fadr, spiReg
	call	spi_write
	movff	fval, spiReg
	call	spi_write
	SPI_OFF
	ENDM	

;SPI write MACRO (adr = register, val = literal)
spi_write_fadr_lval MACRO fadr, lval
    SPI_ON
	movff	fadr, spiReg
	call	spi_write
	movlw	lval
    movwf   spiReg
	call	spi_write
	SPI_OFF
	ENDM	

;SPI write MACRO (adr = literal, val = register)
spi_write_ladr_fval MACRO ladr, fval
    SPI_ON
	movlw   ladr
    movwf   spiReg
	call	spi_write
	movff	fval, spiReg
	call	spi_write
	SPI_OFF
	ENDM	

;SPI write MACRO (adr = literal, val = literal)
spi_write_ladr_lval MACRO ladr, lval
    SPI_ON
	movlw   ladr
    movwf   spiReg
	call	spi_write
	movlw	lval
    movwf   spiReg
	call	spi_write
	SPI_OFF
	ENDM	

;SPI write strobe command (lval = literal)
spi_write_strobe_command MACRO lval
    SPI_ON
	movlw   lval
    movwf   spiReg
	call	spi_write
	SPI_OFF
	ENDM

;SPI read MACRO  (adr = registers)
spi_read_fadr MACRO fadr
    SPI_ON
	movff	fadr, spiReg
	bsf		spiReg, 6
	call	spi_write
    nop
	call	spi_read
	SPI_OFF
	ENDM	

;SPI read MACRO  (adr = literal)
spi_read_ladr MACRO ladr
    SPI_ON
	movlw	ladr
    movwf   spiReg
	bsf		spiReg, 6
	call	spi_write
    nop
	call	spi_read
	SPI_OFF
	ENDM	

;restore FSR0 et FSR1 aprs interrupt
restoreFSR01 MACRO
	movf	FSR0L, w				;FSR0L -> FSR0L_SHAD
	banksel	BSR_SHAD
	movwf	FSR0L_SHAD
	banksel	BSR						;FSR0H -> FSR0H_SHAD
	movf	FSR0H, w
	banksel	BSR_SHAD
	movwf	FSR0H_SHAD
	banksel	BSR						;FSR1L -> FSR1L_SHAD
	movf	FSR1L, w
	banksel	BSR_SHAD
	movwf	FSR1L_SHAD
	banksel	BSR						;FSR1H -> FSR1H_SHAD
	movf	FSR1H, w
	banksel	BSR_SHAD
	movwf	FSR1H_SHAD
	banksel	BSR
	ENDM

;inverse signe de a (5cy)
NEG MACRO aHI, aLO 
	comf 	aLO, f
	comf	aHI, f	
	incf 	aLO, f			
	btfsc 	STATUS, C
	incf 	aHI, f
	ENDM

;ajoute a et b, resultat dans b (6cy)
ADD	MACRO aHI, aLO, bHI, bLO
	movf 	aLO, w			
	addwf 	bLO, f
	movf 	aHI, w
	addwfc 	bHI, f
	ENDM

;DIV2
DIV2 MACRO highByte, lowByte
	bcf		STATUS, C
	rrf		highByte, f
	rrf		lowByte, f
	ENDM

;SUB1500
SUB1500	MACRO  highByte, lowByte
	movlw 	0x24			;low(-1500)
	addwf 	lowByte, f
	movlw 	0xFA			;high(-1500)
	addwfc 	highByte, f
	ENDM

;SUB1000
SUB1000	MACRO  highByte, lowByte
	movlw 	0x18			;low(-1000)
	addwf 	lowByte, f
	movlw 	0xFC			;high(-1000)
	addwfc 	highByte, f
	ENDM

DEADBAND MACRO val			;deadband = 8/250 = about 4%
	movlw	.8				
	subwf	val, f			;val <- val - 8 
	btfss	STATUS, C		;if val >= 0 then skip next instruction 
	clrf	val				;val <- 0
	ENDM

;******************************************************************************
; VARIABLES
;******************************************************************************

    CBLOCK 0x20 ; Sample GPR variable reg allocations

	;init A7105
	cptReg

	;spi_write, spi_read
 	cptBit, spiReg, adressReg, valReg

    ;write_eeprom_pic, read_eeprom_pic
    adr_pic, data_pic

    ;global variables
    channel             ;rx channel no
    cptTimeOut          ;time out counter while waiting for valid bind packed
    id0, id1, id2, id3  ;TX id found in current RX packet
    txid0, txid1, txid2, txid3  ;TX id (current bind or previous bind)

    ip1h, ip1l         ;Input Pulse duration for servo 1 to 8 (high,low)
    ip2h, ip2l         ; = pulse duration read in current RX packet in us
    ip3h, ip3l
    ip4h, ip4l
    ip5h, ip5l
    ip6h, ip6l
    ip7h, ip7l
    ip8h, ip8l
 
	tmph, tmpl			;temporary 2 bytes word
	
    ;op1h, op1l         ;Output Pulse duration for servo 1 to 4 (high,low)
    ;op2h, op2l         ; = servo output pulse duration in us
    ;op3h, op3l
    ;op4h, op4l

    cptChannel          ;RX channel sequence counter (goes from 16 to 1)
    tableOffset         ;offset sequence in tableSequence
	channelOffset		;channel offset (to be subtract to channel number)
	tableAdress 		;for table read using BRW and RETLW

    cptItPWM         	;timer0/PWM it counter
 
	direction						;direction of aileron, elevator, rudder
	motVal, ailVal, eleVal, rudVal	;absolute value of motor, aileron, elevator, rudder
				
	ENDC
;------------------------------------------------------------------------------
; EEPROM INITIALIZATION
;
; The 16F1827 has 256 bytes of non-volatile EEPROM, starting at address 0xF000
; 
;------------------------------------------------------------------------------

DATAEE    ORG  0xF000
    DE    "MCHP"  ; Place 'M' 'C' 'H' 'P' at address 0,1,2,3

;------------------------------------------------------------------------------
; RESET VECTOR
;------------------------------------------------------------------------------

    ORG     0x0000            ; processor reset vector
    PAGESEL START
    GOTO    START             ; When using debug header, first inst.
                              ; may be passed over by ICD2.  

;------------------------------------------------------------------------------
; INTERRUPT SERVICE ROUTINE
;------------------------------------------------------------------------------

    ORG      0x0004

;------------------------------------------------------------------------------
; USER INTERRUPT SERVICE ROUTINE GOES HERE
;------------------------------------------------------------------------------

; Note the 16F1827 family automatically handles context restoration for 
; W, STATUS, BSR, FSR, and PCLATH where previous templates for 16F families
; required manual restoration.  Shadow registers store these SFR values, and
; shadow registers may be modified since they are readable and writable for
; modification to the context restoration.    


	;timer0 interrupt ???
    btfss   INTCON, TMR0IF
    goto    timer1_2_4_6

;*********************************
;timer 0 interrupt (PWM frequency)
;*********************************
timer0_isr
    bcf     INTCON, TMR0IF		;clear timer0 it flag
	banksel	PORTA
    decf    cptItPWM, f			;decremente PWM it counter

	;set timer 1
	banksel	TMR1H
	movlwf	0xFF, TMR1H			;set timer 1 counter registers
	comf	motVal, w
	movwf	TMR1L
	bsf		T1CON, TMR1ON		;start timer 1

	;set timer 2
	movf	ailVal, w
	banksel	TMR2
	clrf	TMR2				;clear timer 2
	movwf	PR2					;load period register
	bsf		T2CON, TMR2ON		;start timer 2

	;set timer 4
	movf	eleVal, w
	banksel	TMR4
	clrf	TMR4				;clear timer 4
	movwf	PR4					;load period register
	bsf		T4CON, TMR4ON		;start timer 4

	;set timer 6
	banksel PORTA
	movf	rudVal, w
	banksel	TMR6
	clrf	TMR6				;clear timer 6
	movwf	PR6					;load period register
	bsf		T6CON, TMR6ON		;start timer 6

	;set motor output
	banksel PORTA
#ifdef RELEASE
	movf	motVal, f			;skip if motor = 0
	btfsc	STATUS, Z
	goto	setAileron
	bsf		MOTOR	
	bsf		MOTOR	
#endif

	;set aileron output
setAileron
	movf	ailVal, f			;skip if aileron = 0
	btfsc	STATUS, Z
	goto	setElevator
	btfsc	ailDir
	goto	ailNeg
ailPos	
	bsf		AIL1
	bsf		AIL1
	bcf		AIL2
	bcf		AIL2
	goto	setElevator
ailNeg	
	bcf		AIL1
	bcf		AIL1
	bsf		AIL2
	bsf		AIL2
	
	;set elevator output
setElevator
	movf	eleVal, f			;skip if elevator = 0
	btfsc	STATUS, Z
	goto	setRudder
	btfsc	eleDir
	goto	eleNeg
elePos
	bsf		ELE1
	bsf		ELE1
	bcf		ELE2
	bcf		ELE2
	goto	setRudder
eleNeg
	bcf		ELE1
	bcf		ELE1
	bsf		ELE2
	bsf		ELE2
	
	;set rudder output
setRudder
	movf	rudVal, f			;skip if rudder = 0
	btfsc	STATUS, Z
	RETFIE 						;quit interrupt, automatic context saving
	btfsc	rudDir
	goto	rudNeg
rudPos
	bsf		RUD1
	bsf		RUD1
	bcf		RUD2	
	bcf		RUD2	
	RETFIE 						;quit interrupt, automatic context saving
rudNeg
	bcf		RUD1
	bcf		RUD1
	bsf		RUD2
	bsf		RUD2
	RETFIE 						;quit interrupt, automatic context saving

	;timer1 interrupt ???
timer1_2_4_6
	banksel	PIR1
    btfss   PIR1, TMR1IF
	goto	timer2_4_6

;*********************************
;timer 1 interrupt (motor pulse)
;*********************************
timer1_isr
	banksel	T1CON
	bcf		T1CON, TMR1ON		;stop timer 1
	banksel PIR1
	bcf		PIR1, TMR1IF		;clear timer1 it flag
	banksel PORTA
#ifdef RELEASE
	bcf		MOTOR				;stop pulse
	bcf		MOTOR
#endif
	RETFIE

	;timer2 interrupt ???
timer2_4_6
	banksel	PIR1
    btfss   PIR1, TMR2IF
	goto	timer4_6

;*********************************
;timer 2 interrupt (aileron pulse)
;*********************************
timer2_isr
	banksel T2CON
	bcf		T2CON, TMR2ON		;stop timer2
	banksel PIR1
    bcf     PIR1, TMR2IF        ;clear timer2 it flag
	banksel PORTA
	bcf		AIL1				;stop pulse ailVal
	bcf		AIL1
	bcf		AIL2
	bcf		AIL2
	RETFIE

	;timer4 interrupt ???
timer4_6
	banksel	PIR3
    btfss   PIR3, TMR4IF
	goto	timer6_isr

;*********************************
;timer 4 interrupt (elevator pulse)
;*********************************
timer4_isr
	banksel T4CON
	bcf		T4CON, TMR4ON		;stop timer4
	banksel PIR3
    bcf     PIR3, TMR4IF        ;clear timer4 it flag
	banksel PORTA
	bcf		ELE1				;stop pulse eleVal
	bcf		ELE1
	bcf		ELE2
	bcf		ELE2
	RETFIE

;*********************************
;timer 6 interrupt (rudVal pulse)
;*********************************
timer6_isr
	banksel T6CON
	bcf		T6CON, TMR6ON		;stop timer6
	banksel PIR3
    bcf     PIR3, TMR6IF        ;clear timer6 it flag
	banksel PORTA
	bcf		RUD1				;stop pulse rudVal
	bcf		RUD1
	bcf		RUD2
	bcf		RUD2
	RETFIE


;******************************************************
;write_eeprom_pic
;******************************************************
write_eeprom_pic
	banksel	EECON1				;bank 3
    bcf     EECON1, CFGS        ;accs  EEPROM ou program
    bcf     EECON1, EEPGD       ;accs  EEPROM
	banksel	PIR2				;bank 0
    bcf     PIR2, EEIF          ;efface flag fin d'criture
	movf	data_pic, w
	banksel	EECON1				;bank 3
	movwf	EEDATL
	banksel	PORTA				;bank 0
	movf	adr_pic, w
	banksel	EECON1				;bank 3
	movwf	EEADR
    bsf     EECON1, WREN        ;Enable write
    movlw   0x55                ;Unlock write
    movwf   EECON2 ;
    movlw   0xAA ;
    movwf   EECON2 ;
    bsf     EECON1, WR          ;Start the write
    ;attente fin ecriture
	banksel	PIR2				;bank 0
attenteFinEcriture
    btfss   PIR2, EEIF
    goto    attenteFinEcriture
	banksel	EECON1
    bcf     EECON1, WREN         ;dsactive write
	banksel	PORTA
    return

;******************************************************
;read_eeprom_pic
;******************************************************
read_eeprom_pic
	banksel	EECON1				;bank 3
    bcf     EECON1, CFGS        ;accs  EEPROM ou program
    bcf     EECON1, EEPGD       ;accs  EEPROM
	banksel	PORTA				;bank 0
    movf   	adr_pic, w
	banksel	EECON1				;bank 3
	movwf	EEADR
    bsf     EECON1, RD          ;EE Read
	movf	EEDATL, w
	banksel	PORTA				;bank 0
    movwf	data_pic    		;Move data to data_pic
    return

;****************************************
; spi_write
; entre:
;   spiReg: valeur du registre  crire
;****************************************
spi_write
	movlwf   .8, cptBit
	banksel	PORTA
boucleWriteReg
    btfss   spiReg, 7
    bcf     SDIO_SPI
    btfsc   spiReg, 7
    bsf     SDIO_SPI
	nop
    SCK_SPI_PULSE
    rlf    	spiReg, f
    decfsz  cptBit, f
    goto    boucleWriteReg;
    return

;*********************************
; spi_read
; sortie:
;   spiReg: valeur du registre lu
;*********************************
spi_read
	banksel	TRISA
    bsf     SDIO_SPI_TRIS       ;SDIO input
	banksel	PORTA
    movlwf   .8, cptBit
    clrf    spiReg
    bcf     STATUS, C
boucleReadReg
    rlf     spiReg, f
	btfsc   SDIO_SPI
    bsf     spiReg, 0
    SCK_SPI_PULSE
    decfsz  cptBit, f
    goto    boucleReadReg
	banksel	TRISA
    bcf     SDIO_SPI_TRIS       ;SDIO output
	banksel	PORTA
    return

;**************
; read_packet
;**************
read_packet
    SPI_ON
	movlwf	0x45, spiReg
	call	spi_write
	call	spi_read        ;sync
	call	spi_read        ;id0 TX (lsb)
    movff   spiReg, id0
	call	spi_read        ;id1 TX
    movff   spiReg, id1
	call	spi_read        ;id2 TX
    movff   spiReg, id2
	call	spi_read        ;id3 TX (msb)
    movff   spiReg, id3
	call	spi_read        ;pulse duration 1
    movff   spiReg, ip1l
	call	spi_read
    movff   spiReg, ip1h
	call	spi_read        ;pulse duration 2
    movff   spiReg, ip2l
	call	spi_read
    movff   spiReg, ip2h
	call	spi_read        ;pulse duration 3
    movff   spiReg, ip3l
	call	spi_read
    movff   spiReg, ip3h
	call	spi_read        ;pulse duration 4
    movff   spiReg, ip4l
	call	spi_read
    movff   spiReg, ip4h
	call	spi_read        ;pulse duration 5
    movff   spiReg, ip5l
	call	spi_read
    movff   spiReg, ip5h
	call	spi_read        ;pulse duration 6
    movff   spiReg, ip6l
	call	spi_read
    movff   spiReg, ip6h
	call	spi_read        ;pulse duration 7
    movff   spiReg, ip7l
	call	spi_read
    movff   spiReg, ip7h
	call	spi_read        ;pulse duration 8
    movff   spiReg, ip8l
	call	spi_read
    movff   spiReg, ip8h
    SPI_OFF
    return

;**************
; init_A7105
;**************
init_A7105

    ;reset A7105
	banksel	PORTA
	spi_write_ladr_lval 0x00, 0x00

	;ID A7105 <- 0x5475c52a ??? useful ???
	SPI_ON
	movlwf	0x06, spiReg
	call	spi_write
	movlwf	0x54, spiReg
	call	spi_write
	movlwf	0x75, spiReg
	call	spi_write
	movlwf	0xc5, spiReg
	call	spi_write
	movlwf	0x2A, spiReg
	call	spi_write
	SPI_OFF

	;read ID A7105 -> id0..id3 to check spi dialog
	SPI_ON
	movlwf	0x46, spiReg
	call	spi_write
	call	spi_read
	movff	spiReg, id0
	call	spi_read
	movff	spiReg, id1
	call	spi_read
	movff	spiReg, id2
	call	spi_read
	movff	spiReg, id3
	SPI_OFF


    ;init A7105 regs
	clrf	tableAdress			;reset table adress
	movlwf	0xFF, adressReg		;init reg adress = -1
	movlwf	0x33, cptReg		;init loop counter
BoucleInitReg
	incf	adressReg, f		;increment reg adress
	movf	adressReg, w		;init w pour "brw" (relative branch with w)
	call	tableInitReg		;read byte in tableInitReg -> valReg
	movwf	valReg
	incf	tableAdress, f		;increment table adress
	incf	valReg, w			;test valReg == 0xFF ?
	bz		BoucleInitReg		;yes -> skip reg init
	spi_write_fadr_fval adressReg, valReg
	decfsz	cptReg, f			;test loop end ?
	goto	BoucleInitReg		;no -> loop

	;IF Filter Bank Calibration
	spi_write_ladr_lval 0x02, 0x01
waitCalibrationFilter			;wait calibration end
	spi_read_ladr	0x02
	btfsc	spiReg, 0
	goto	waitCalibrationFilter

	;test calibration ok ?

	;VCO Current Calibration
	spi_write_ladr_lval 0x24, 0x013

	;VCO Bank Calibration
	spi_write_ladr_lval 0x26, 0x03B

    ;VCO Bank Calibrate channel 0x00?
	spi_write_ladr_lval 0x0F, 0x00	;Set Channel 0
    spi_write_ladr_lval 0x02, 0x02	;VCO Calibration
waitCalibrationVCO					;wait calibration end
	spi_read_ladr	0x02
	btfsc	spiReg, 1
	goto	waitCalibrationVCO

	;test calibration ok ?

    ;VCO Bank Calibrate channel 0xa0?
	spi_write_ladr_lval 0x0F, 0xA0	;Set Channel 0xa0
    spi_write_ladr_lval 0x02, 0x02	;VCO Calibration
waitCalibrationVCObis				;wait calibration end
	spi_read_ladr	0x02
	btfsc	spiReg, 1
	goto	waitCalibrationVCObis

	;test calibration ok ?

    ;Reset VCO Band calibration
    spi_write_ladr_lval 0x25, 0x08

    return

;**************
; bind_TX
;*************

bind_TX

    ;wait for valid bind packet for about 2,5s on channel 0
    ;(256 test loop max, 10ms each loop)
    clrf    cptTimeOut                  ;init time out counter

waitValidBindPacket
    spi_write_strobe_command 0xA0       ;A0 -> standby mode
    spi_write_strobe_command 0xF0       ;F0 -> reset RX FIFO
    spi_write_ladr_lval 0x0F, 0x00      ;Set Channel 0
    spi_write_strobe_command 0xC0       ;C0 -> get into RX mode

waitBindPacketOrTimeOut
	banksel	PORTA             			;wait for 10ms (39x256us)
    movlw	.39				
	call	waitN256us

#ifdef RELEASE
    btfsc   cptTimeOut, 2               ;12Hz blinking LED while binding
    bcf     LED                         ;LED OFF
    btfss   cptTimeOut, 2
    bsf     LED                         ;LED ON
#endif

    btfss   GIO                         ;received RX packet ?
    goto    testValidBindPacket         ;yes -> test valid packet ?

    decfsz  cptTimeOut, f               ;decremente time out counter, roll-off?
    goto    waitBindPacketOrTimeOut     ;no  -> wait again
    goto    timeOutBindPacket           ;yes -> time out

testValidBindPacket
    spi_read_ladr	0x00                ;read mode register
    btfsc   spiReg, 5                   ;test CRC (bit5)
    goto    waitValidBindPacket         ;not valid, wait again
    btfsc   spiReg, 6                   ;test CEF (bit6)
    goto    waitValidBindPacket         ;not valid, wait again

    ;a valid bind packet was received -> write 4 bytes ID
    ;in PIC EEPROM to avoid binding each time RX is switched on
    call    read_packet                 ;read packet which contains TX ID
    clrf    adr_pic
    movff   id0, data_pic               ;write id0, eeprom adress 0x00
    call    write_eeprom_pic
    incf    adr_pic, f
    movff   id1, data_pic               ;write id1, eeprom adress 0x01
    call    write_eeprom_pic
    incf    adr_pic, f
    movff   id2, data_pic               ;write id2, eeprom adress 0x02
    call    write_eeprom_pic
    incf    adr_pic, f
    movff   id3, data_pic               ;write id3, eeprom adress 0x03
    call    write_eeprom_pic
    movff   id0, txid0                  ;tx id = current id
    movff   id1, txid1
    movff   id2, txid2
    movff   id3, txid3
    return

timeOutBindPacket
    ;could not bind to TX => read previous TX ID stored in PIC EEPROM
    clrf    adr_pic                     ;read id0 @ 0x00
    call    read_eeprom_pic
    movff   data_pic, txid0
    incf    adr_pic, f                  ;read id1 @ 0x01
    call    read_eeprom_pic
    movff   data_pic, txid1
    incf    adr_pic, f                  ;read id2 @ 0x02
    call    read_eeprom_pic
    movff   data_pic, txid2
    incf    adr_pic, f                  ;read id3 @ 0x03
    call    read_eeprom_pic
    movff   data_pic, txid3
    return

;***********************
; readNextChannelNumber
;***********************
readNextChannelNumber
    decfsz  cptChannel, f				;decrement channel counter
    goto    tableReadChannel            ;cptChannel > 0 => skip sequence reset
resetChannelSequence
    movlwf  .16, cptChannel             ;cptChannel = 16
    movff    tableOffset, tableAdress	;load adress of tableChannels
tableReadChannel
	movf	tableAdress, w
    call	tableChannels               ;read channel number in table
    movwf   channel
	decf	channel, f                  ;dec channel number
	movf	channelOffset, w			;substract channel offset from channel number
	subwf	channel, f
	incf	tableAdress, f				;incremente table adress
	return

;******************************
; waitN256us: wait (N x 256us)
;******************************
waitN256us
	movwf	cptItPWM
wait
	movf	cptItPWM, f
	btfss	STATUS, Z
	goto	wait
	return

;------------------------------------------------------------------------------
; MAIN PROGRAM
;------------------------------------------------------------------------------

START

;------------------------------------------------------------------------------
; PLACE USER PROGRAM HERE
;------------------------------------------------------------------------------

    bcf		INTCON, GIE		; disable global interrupt

	;set internal oscillator speed Fosc=16Mhz
    ;=> internal instruction cycle speed = Fosc/4 = 4Mhz
	banksel OSCCON
    bsf		OSCCON, IRCF3
    bsf		OSCCON, IRCF2
    bsf		OSCCON, IRCF1
    bsf		OSCCON, IRCF0

	;disable global interrupt
	bsf		INTCON, GIE

    ;disable analog inputs to enable digital inputs
	banksel	ANSELA
    clrf    ANSELA
    clrf    ANSELC

	;config TRIS A,B,C
	banksel	TRISA
    bcf     CS_SPI_TRIS         ;CS digital output
    bcf     SDIO_SPI_TRIS       ;SDIO digital output
    bcf     SCK_SPI_TRIS        ;SCK digital output
#ifdef RELEASE
    bcf     LED_TRIS			;LED digital output
    bcf     MOTOR_TRIS			;motor digital output
#endif
	bcf		AIL1_TRIS			;aileron digital output
	bcf		AIL2_TRIS
	bcf		ELE1_TRIS			;elevator digital output
	bcf		ELE2_TRIS
	bcf		RUD1_TRIS			;rudder digital output
	bcf		RUD2_TRIS

    ;config PORT A,B,C
	banksel	PORTA
    clrf    PORTA
    clrf    PORTC
#ifdef RELEASE
    bsf     LED                 ;LED ON
#endif
    bsf     CS_SPI              ;CS default HIGH
   	bsf     SDIO_SPI            ;SDIO default HIGH
    bcf     SCK_SPI             ;SCK  default LOW

	;set timer0 ( set PWM frequency )
    ;8 bit timer
	banksel OPTION_REG
	clrf	OPTION_REG		; clk = Fosc/4 = 4Mhz, enable prescaler, 
	bsf		OPTION_REG, PS0	; PSPS<2:0> = 001 = 1:4 -> F = 4E6/4/256 = 3906.25Hz
    bcf     INTCON, TMR0IF  ; clear flag interrupt timer0
	bsf		INTCON, TMR0IE	; enable interrupt timer0

	;set timer1 ( set motor pulse duration )
    ;16 bit timer
	banksel T1CON
    clrf    T1CON			; reset timer1 control register, timer1 stopped, clk = FOSC/4
    clrf    T1GCON			; reset timer1 gate control register
    bsf     T1CON, T1CKPS1  ; prescaler 1:4 (1us period)
	bcf		PIR1, TMR1IF	; clear timer1 it flag
	banksel	PIE1
	bsf		PIE1, TMR1IE	; enable interrupt timer1

    ;set timer2 ( set aileron pulse duration )
    ;8 bits timer with comparator register
	banksel TMR2
 	clrf	PR2				; reset comparator register
	clrf	T2CON			; reset control register, 1:1 postscaler, timer off 
	bsf		T2CON, T2CKPS0  ; prescaler 1:4 (1us period)
	banksel	PIR1
	bcf		PIR1, TMR2IF	; clear flag interrupt timer2
	banksel	PIE1
	bsf		PIE1, TMR2IE	; enable interrupt timer2

    ;set timer4 ( set elevator pulse duration )
    ;8 bits timer with comparator register
	banksel TMR4
 	clrf	PR4				; reset comparator register
	clrf	T4CON			; reset control register, 1:1 postscaler, timer off 
	bsf		T4CON, T4CKPS0  ; prescaler 1:4 (1us period)
	banksel	PIR3
	bcf		PIR3, TMR4IF	; clear flag interrupt timer4
 	banksel	PIE3
	bsf		PIE3, TMR4IE	; enable interrupt timer4

    ;set timer6 ( set rudder pulse duration )
    ;8 bits timer with comparator register
	banksel TMR6
 	clrf	PR6				; reset comparator register
	clrf	T6CON			; reset control register, 1:1 postscaler, timer off 
	bsf		T6CON, T6CKPS0  ; prescaler 1:4 (1us period)
	banksel	PIR3
	bcf		PIR3, TMR6IF	; clear flag interrupt timer6
	banksel	PIE3
	bsf		PIE3, TMR6IE	; enable interrupt timer6

	;init absolute value for motor, aileron, elevator, rudder
	banksel PORTA
	movlwf	.0, motVal
	movlwf	.0, ailVal
	movlwf	.0, eleVal
	movlwf	.0, rudVal

	;init direction for aileron, elevator, rudder
	bsf		ailDir
	bcf		eleDir
	bcf		rudDir

	;debug: apply DEADBAND to aileron, elevator, rudder
	;DEADBAND ailVal
	;DEADBAND eleVal
	;DEADBAND rudVal

	;enable interrupts for all timers
	bsf		INTCON, PEIE	; enable peripheral interrupt
	bsf		INTCON, GIE		; enable global interrupt

	;wait about 10ms to be sure A7105 wake up
    movlw	.39				;39x256us
	call	waitN256us

	;init A7105
    call    init_A7105
    nop

    ;bind with TX
    call    bind_TX

    ;------------------------------------------------------------------

    ;now we have TX ID (previous TX or just binded TX)
    ;we read each packet in a frequency sequence which depends on TX id
    ;(frequency sequences are in tableChannels)

    swapf   txid0, w			;init table read channel sequence
    andlw   0xF0            	;compute tableOffset <- (id0 & 0x0F)*16
    movwf   tableOffset

    movlwf  0x01, cptChannel    ;init channel counter (goes from 16 to 1)

	swapf	txid0, w			;compute channel offset
    andlw   0x0F            	;channelOffset <- (id0 / 16)
	movwf	channelOffset
	movlw	.10					;if channelOffset >= 10 then channelOffset <- 9
	subwf	channelOffset, w
	movlw	.9
	btfsc	STATUS, C			;(C=1 if W<=F)
	movwf	channelOffset

    ;***********
    ; MAIN LOOP
    ;***********

scanChannels

	call readNextChannelNumber			;read next channel number in tableChannels

    spi_write_strobe_command 0xA0       ;A0 -> standby mode
    spi_write_strobe_command 0xF0       ;F0 -> reset RX FIFO
    spi_write_ladr_fval 0x0F, channel	;set channel
    spi_write_strobe_command 0xC0       ;C0 -> get into RX mode

    movlw	.8			                ;set cptItPWM for about 2ms count down (8x256us)
	movwf	cptItPWM

waitRxPacket
	banksel	PIR3
    movf	cptItPWM, f	                ;test time-out 2ms ?
	btfsc	STATUS, Z
	goto	timeOutRxPacket

	banksel	PORTA
    btfsc   GIO                         ;test packet available ?
    goto    waitRxPacket                ;no -> wait

    spi_read_ladr	0x00                ;read "mode register"
    btfsc   spiReg, 5                   ;test CRC (bit5)
    goto    waitRxPacket                ;not valid, wait again
    btfsc   spiReg, 6                   ;test CEF (bit6)
    goto    waitRxPacket                ;not valid, wait again
    call    read_packet                 ;read RX packet
    movf    txid0, w                    ;good TX ID0 ?
    cpfseq  id0
    goto    waitRxPacket
    movf    txid1, w                    ;good TX ID1 ?
    cpfseq  id1
    goto    waitRxPacket
    movf    txid2, w                    ;good TX ID2 ?
    cpfseq  id2
    goto    waitRxPacket
    movf    txid3, w                    ;good TX ID3 ?
    cpfseq  id3
    goto    waitRxPacket

	banksel	PORTA
#ifdef RELEASE
    bsf     LED							;led ON = rx ok
#endif

	;update pwm values for motor
	movff	ip3h, tmph		; tmp <- ip1 - 1000
	movff	ip3l, tmpl
	SUB1000 tmph, tmpl		
	btfss	tmph, 7			; tmp < 0 ?
	goto	motorPos
	clrf	tmph			; yes: tmp <- 0
	clrf	tmpl
motorPos
	DIV2	tmph, tmpl		; tmp range [0..1000] -> [0..250]
	DIV2	tmph, tmpl
	movff	tmpl, motVal	; motVal <- tmp

	;update pwm values for aileron
	movff	ip1h, tmph		; tmp <- ip1 - 1500
	movff	ip1l, tmpl
	SUB1500 tmph, tmpl	
	bcf		ailDir			; clear flag direction	(default value)
	btfss	tmph, 7			; tmp < 0 ?
	goto	aileronPos
	NEG		tmph, tmpl		; yes: tmp <- |tmp|, set flag direction
	bsf		ailDir
aileronPos
	DIV2	tmph, tmpl		; tmp range [0..500] -> [0..250]
	DEADBAND tmpl			; apply deadband
	movff	tmpl, ailVal	; ailVal <- tmp

	;update pwm values for elevator
	movff	ip2h, tmph		; tmp <- ip1 - 1500
	movff	ip2l, tmpl
	SUB1500 tmph, tmpl	
	bcf		eleDir			; clear flag direction	(default value)
	btfss	tmph, 7			; tmp < 0 ?
	goto	elevatorPos
	NEG		tmph, tmpl		; yes: tmp <- |tmp|, set flag direction
	bsf		eleDir	
elevatorPos
	DIV2	tmph, tmpl		; tmp range [0..500] -> [0..250]
	DEADBAND tmpl			; apply deadband
	movff	tmpl, eleVal	; eleVal <- tmp

	;update pwm values for rudder
	movff	ip4h, tmph		; tmp <- ip1 - 1500
	movff	ip4l, tmpl
	SUB1500 tmph, tmpl	
	bcf		rudDir			; clear flag direction	(default value)
	btfss	tmph, 7			; tmp < 0 ?
	goto	rudderPos
	NEG		tmph, tmpl		; yes: tmp <- |tmp|, set flag direction
	bsf		rudDir	
rudderPos
	DIV2	tmph, tmpl		; tmp range [0..500] -> [0..250]
	DEADBAND tmpl			; apply deadband
	movff	tmpl, rudVal	; rudVal <- tmp

	;next channel...
    goto    scanChannels

timeOutRxPacket
	banksel	PORTA
#ifdef RELEASE
	bcf		LED							;switch off LED
#endif
	call	readNextChannelNumber		;skip next channel because we probably allready missed it
 	goto	scanChannels				;scan next channel

;**********************************************************
;*********************** TABLES ***************************
;**********************************************************

;A7105 config registers init value
tableInitReg
	brw
	retlw 0xFF
	retlw 0x42
	retlw 0x00
	retlw 0x14
	retlw 0x00
	retlw 0xFF
	retlw 0xFF 
	retlw 0x00
	retlw 0x00
	retlw 0x00
	retlw 0x00
	retlw 0x01
	retlw 0x21
	retlw 0x05
	retlw 0x00
	retlw 0x50
	retlw 0x9e
	retlw 0x4b
	retlw 0x00
	retlw 0x02
	retlw 0x16
	retlw 0x2b
	retlw 0x12
	retlw 0x00
	retlw 0x62
	retlw 0x80
	retlw 0x80
	retlw 0x00
	retlw 0x0a
	retlw 0x32
	retlw 0xc3
	retlw 0x0f
	retlw 0x13
	retlw 0xc3
	retlw 0x00
	retlw 0xFF
	retlw 0x00
	retlw 0x00
	retlw 0x3b
	retlw 0x00
	retlw 0x17
	retlw 0x47
	retlw 0x80
	retlw 0x03
	retlw 0x01
	retlw 0x45
	retlw 0x18
	retlw 0x00
	retlw 0x01
	retlw 0x0f
	retlw 0xFF

    ;substract one from each tableChannels value to have good channels numbers ...
tableChannels
	brw
	;line 0
	retlw 0x0a
	retlw 0x5a
	retlw 0x14
	retlw 0x64
	retlw 0x1e
	retlw 0x6e
	retlw 0x28
	retlw 0x78
	retlw 0x32
	retlw 0x82
	retlw 0x3c
	retlw 0x8c
	retlw 0x46
	retlw 0x96
	retlw 0x50
	retlw 0xa0
	
	;line 1
	retlw 0xa0
	retlw 0x50
	retlw 0x96
	retlw 0x46
	retlw 0x8c
	retlw 0x3c
	retlw 0x82
	retlw 0x32
	retlw 0x78
	retlw 0x28
 	retlw 0x6e
	retlw 0x1e
	retlw 0x64
	retlw 0x14
	retlw 0x5a
	retlw 0x0a
	
	;line 2
	retlw 0x0a
	retlw 0x5a
	retlw 0x50
	retlw 0xa0
	retlw 0x14
	retlw 0x64
	retlw 0x46
	retlw 0x96
	retlw 0x1e
	retlw 0x6e
	retlw 0x3c
	retlw 0x8c
	retlw 0x28
	retlw 0x78
	retlw 0x32
	retlw 0x82
	
	;line 3
	retlw 0x82
	retlw 0x32
	retlw 0x78
	retlw 0x28
	retlw 0x8c
	retlw 0x3c
	retlw 0x6e
	retlw 0x1e
	retlw 0x96
	retlw 0x46
	retlw 0x64
	retlw 0x14
	retlw 0xa0
	retlw 0x50
	retlw 0x5a
	retlw 0x0a
	
	;line 4
	retlw 0x28
	retlw 0x78
	retlw 0x0a
	retlw 0x5a
	retlw 0x50
	retlw 0xa0
	retlw 0x14
	retlw 0x64
	retlw 0x1e
	retlw 0x6e
	retlw 0x3c
	retlw 0x8c
	retlw 0x32
	retlw 0x82
	retlw 0x46
	retlw 0x96
	
	;line 5
	retlw 0x96
	retlw 0x46
	retlw 0x82
	retlw 0x32
	retlw 0x8c
	retlw 0x3c
	retlw 0x6e
	retlw 0x1e
	retlw 0x64
	retlw 0x14
	retlw 0xa0
	retlw 0x50
	retlw 0x5a
	retlw 0x0a
	retlw 0x78
	retlw 0x28
	
	;line 6
	retlw 0x50
	retlw 0xa0
	retlw 0x28
	retlw 0x78
	retlw 0x0a
	retlw 0x5a
	retlw 0x1e
	retlw 0x6e
	retlw 0x3c
	retlw 0x8c
	retlw 0x32
	retlw 0x82
	retlw 0x46
	retlw 0x96
	retlw 0x14
	retlw 0x64
	
	;line 7
	retlw 0x64
	retlw 0x14
	retlw 0x96
	retlw 0x46
	retlw 0x82
	retlw 0x32
	retlw 0x8c
	retlw 0x3c
	retlw 0x6e
	retlw 0x1e
	retlw 0x5a
	retlw 0x0a
	retlw 0x78
	retlw 0x28
	retlw 0xa0
	retlw 0x50
	
	;line 8
	retlw 0x50
	retlw 0xa0
	retlw 0x46
	retlw 0x96
	retlw 0x3c
	retlw 0x8c
	retlw 0x28
	retlw 0x78
	retlw 0x0a
	retlw 0x5a
	retlw 0x32
	retlw 0x82
	retlw 0x1e
	retlw 0x6e
	retlw 0x14
	retlw 0x64
	
	;line 9
	retlw 0x64
	retlw 0x14
	retlw 0x6e
	retlw 0x1e
	retlw 0x82
	retlw 0x32
	retlw 0x5a
	retlw 0x0a
	retlw 0x78
	retlw 0x28
	retlw 0x8c
	retlw 0x3c
	retlw 0x96
	retlw 0x46
	retlw 0xa0
	retlw 0x50
	
	;line 10
	retlw 0x46
	retlw 0x96
	retlw 0x3c
	retlw 0x8c
	retlw 0x50
	retlw 0xa0
	retlw 0x28
	retlw 0x78
	retlw 0x0a
	retlw 0x5a
	retlw 0x1e
	retlw 0x6e
	retlw 0x32
	retlw 0x82
	retlw 0x14
	retlw 0x64
	
	;line 11
	retlw 0x64
	retlw 0x14
	retlw 0x82
	retlw 0x32
	retlw 0x6e
	retlw 0x1e
	retlw 0x5a
	retlw 0x0a
	retlw 0x78
	retlw 0x28
	retlw 0xa0
	retlw 0x50
	retlw 0x8c
	retlw 0x3c
	retlw 0x96
	retlw 0x46
	
	;line 12
	retlw 0x46
	retlw 0x96
	retlw 0x0a
	retlw 0x5a
	retlw 0x3c
	retlw 0x8c
	retlw 0x14
	retlw 0x64
	retlw 0x50
	retlw 0xa0
	retlw 0x28
	retlw 0x78
	retlw 0x1e
	retlw 0x6e
	retlw 0x32
	retlw 0x82
	
	;line 13
	retlw 0x82
	retlw 0x32
	retlw 0x6e
	retlw 0x1e
	retlw 0x78
	retlw 0x28
	retlw 0xa0
	retlw 0x50
	retlw 0x64
	retlw 0x14
	retlw 0x8c
	retlw 0x3c
	retlw 0x5a
	retlw 0x0a
	retlw 0x96
	retlw 0x46
	
	;line 14
	retlw 0x46
	retlw 0x96
	retlw 0x0a
	retlw 0x5a
	retlw 0x50
	retlw 0xa0
	retlw 0x3c
	retlw 0x8c
	retlw 0x28
	retlw 0x78
	retlw 0x1e
	retlw 0x6e
	retlw 0x32
	retlw 0x82
	retlw 0x14
	retlw 0x64
	
	;line 15
	retlw 0x64
	retlw 0x14
	retlw 0x82
	retlw 0x32
	retlw 0x6e
	retlw 0x1e
	retlw 0x78
	retlw 0x28
	retlw 0x8c
	retlw 0x3c
	retlw 0xa0
	retlw 0x50
	retlw 0x5a
	retlw 0x0a
	retlw 0x96
	retlw 0x46

;*********************** F I N **********************

    END